from __future__ import annotations

from importlib.util import find_spec
from typing import TYPE_CHECKING, Optional

import pytest
from pytest import FixtureRequest
from pytest_mock import MockerFixture

from .constants import CLIENT_ID, ENV


if find_spec("starlette") is None:
    pytest.skip("starlette is not available", allow_module_level=True)
else:
    # Need to import BackgroundTasks at package level to avoid NameError in FastAPI
    from starlette.background import BackgroundTasks
    from starlette.requests import Request

if TYPE_CHECKING:
    from starlette.applications import Starlette


@pytest.fixture(
    scope="module",
    params=["starlette", "fastapi"] if find_spec("fastapi") is not None else ["starlette"],
)
async def app(request: FixtureRequest, module_mocker: MockerFixture) -> Starlette:
    module_mocker.patch("apitally.client.client_asyncio.ApitallyClient._instance", None)
    module_mocker.patch("apitally.client.client_asyncio.ApitallyClient.start_sync_loop")
    module_mocker.patch("apitally.client.client_asyncio.ApitallyClient.handle_shutdown")
    if request.param == "starlette":
        return get_starlette_app()
    elif request.param == "fastapi":
        return get_fastapi_app()
    raise NotImplementedError


def get_starlette_app() -> Starlette:
    from starlette.applications import Starlette
    from starlette.responses import PlainTextResponse, StreamingResponse
    from starlette.routing import Mount, Route

    from apitally.starlette import ApitallyConsumer, ApitallyMiddleware, RequestLoggingConfig

    def foo(request: Request):
        request.state.apitally_consumer = "test"
        return PlainTextResponse("foo")

    def foo_bar(request: Request):
        return PlainTextResponse(f"foo: {request.path_params['bar']}")

    async def bar(request: Request):
        body = await request.body()
        return PlainTextResponse("bar: " + body.decode())

    def baz(request: Request):
        raise ValueError("baz")

    def val(request: Request):
        return PlainTextResponse("validation error", status_code=422)

    def stream(request: Request):
        def stream_response():
            yield b"foo"
            yield b"bar"

        return StreamingResponse(stream_response())

    def task(request: Request):
        def task_func_with_error():
            raise ValueError("task")

        tasks = BackgroundTasks()
        tasks.add_task(task_func_with_error)
        return PlainTextResponse("ok", background=tasks)

    def identify_consumer(request: Request) -> Optional[ApitallyConsumer]:
        return ApitallyConsumer("test", name="Test")

    sub_app = Starlette(
        routes=[
            Route("/foo", foo),
            Route("/foo/{bar}", foo_bar),
            Route("/bar", bar, methods=["POST"]),
            Route("/baz", baz, methods=["POST"]),
            Route("/val", val),
        ]
    )
    app = Starlette(
        routes=[
            Mount("/api", sub_app),
            Mount(
                "/test",
                routes=[
                    Route("/task", task, methods=["POST"]),
                ],
            ),
            Route("/stream", stream),
        ]
    )
    app.add_middleware(
        ApitallyMiddleware,
        client_id=CLIENT_ID,
        env=ENV,
        request_logging_config=RequestLoggingConfig(
            enabled=True,
            log_request_body=True,
            log_response_body=True,
        ),
        identify_consumer_callback=identify_consumer,
    )
    return app


def get_fastapi_app() -> Starlette:
    from fastapi import APIRouter, FastAPI, Query
    from fastapi.responses import PlainTextResponse, StreamingResponse

    from apitally.fastapi import ApitallyConsumer, ApitallyMiddleware, RequestLoggingConfig

    def identify_consumer(request: Request) -> Optional[ApitallyConsumer]:
        return ApitallyConsumer("test", name="Test")

    app = FastAPI(title="Test App", description="A simple test app.", version="1.2.3")
    app.add_middleware(
        ApitallyMiddleware,
        client_id=CLIENT_ID,
        env=ENV,
        request_logging_config=RequestLoggingConfig(
            enabled=True,
            log_request_body=True,
            log_response_body=True,
        ),
        identify_consumer_callback=identify_consumer,
    )

    router = APIRouter()

    @router.get("/foo")
    def foo():
        return "foo"

    @router.get("/foo/{bar}")
    def foo_bar(bar: str):
        return PlainTextResponse(f"foo: {bar}")

    @router.post("/bar")
    async def bar(request: Request):
        body = await request.body()
        return PlainTextResponse("bar: " + body.decode())

    @router.post("/baz")
    def baz():
        raise ValueError("baz")

    @router.get("/val")
    def val(foo: int = Query()):
        return "val"

    @app.get("/stream")
    def stream():
        def stream_response():
            yield b"foo"
            yield b"bar"

        return StreamingResponse(stream_response())

    @app.post("/test/task")
    def task(background_tasks: BackgroundTasks):
        def task_func_with_error():
            raise ValueError("task")

        background_tasks.add_task(task_func_with_error)
        return "ok"

    app.include_router(router, prefix="/api")

    return app


def test_middleware_requests_ok(app: Starlette, mocker: MockerFixture):
    from starlette.testclient import TestClient

    mock = mocker.patch("apitally.client.requests.RequestCounter.add_request")

    with TestClient(app) as client:
        response = client.get("/api/foo")
        assert response.status_code == 200
        mock.assert_called_once()
        assert mock.call_args is not None
        assert mock.call_args.kwargs["consumer"] == "test"
        assert mock.call_args.kwargs["method"] == "GET"
        assert mock.call_args.kwargs["path"] == "/api/foo"
        assert mock.call_args.kwargs["status_code"] == 200
        assert mock.call_args.kwargs["response_time"] > 0

        response = client.get("/api/foo/123")
        assert response.status_code == 200
        assert mock.call_count == 2
        assert mock.call_args is not None
        assert mock.call_args.kwargs["path"] == "/api/foo/{bar}"

        response = client.post("/api/bar")
        assert response.status_code == 200
        assert mock.call_count == 3
        assert mock.call_args is not None
        assert mock.call_args.kwargs["method"] == "POST"

        response = client.get("/stream")
        assert response.status_code == 200
        assert mock.call_count == 4
        assert mock.call_args is not None
        assert mock.call_args.kwargs["response_size"] == 6


def test_middleware_requests_error(app: Starlette, mocker: MockerFixture):
    from starlette.testclient import TestClient

    mock1 = mocker.patch("apitally.client.requests.RequestCounter.add_request")
    mock2 = mocker.patch("apitally.client.server_errors.ServerErrorCounter.add_server_error")

    with TestClient(app, raise_server_exceptions=False) as client:
        response = client.post("/api/baz")
        assert response.status_code == 500
        mock1.assert_called_once()
        assert mock1.call_args is not None
        assert mock1.call_args.kwargs["method"] == "POST"
        assert mock1.call_args.kwargs["path"] == "/api/baz"
        assert mock1.call_args.kwargs["status_code"] == 500
        assert mock1.call_args.kwargs["response_time"] > 0

        mock2.assert_called_once()
        assert mock2.call_args is not None
        exception = mock2.call_args.kwargs["exception"]
        assert isinstance(exception, ValueError)

        # Throws a ValueError in a background task, but returns 200
        response = client.post("/test/task")
        assert response.status_code == 200
        assert mock1.call_count == 2
        assert mock1.call_args is not None
        assert mock1.call_args.kwargs["status_code"] == 200
        mock2.assert_called_once()  # Not called again


def test_middleware_requests_unhandled(app: Starlette, mocker: MockerFixture):
    from starlette.testclient import TestClient

    mock = mocker.patch("apitally.client.requests.RequestCounter.add_request")

    with TestClient(app) as client:
        response = client.post("/xxx")
        assert response.status_code == 404
        mock.assert_not_called()


def test_middleware_validation_error(app: Starlette, mocker: MockerFixture):
    from starlette.testclient import TestClient

    mock = mocker.patch("apitally.client.validation_errors.ValidationErrorCounter.add_validation_errors")

    with TestClient(app) as client:
        # Validation error as foo must be an integer
        response = client.get("/api/val?foo=bar")
        assert response.status_code == 422

        # FastAPI only
        if response.headers["Content-Type"] == "application/json":
            mock.assert_called_once()
            assert mock.call_args is not None
            assert mock.call_args.kwargs["method"] == "GET"
            assert mock.call_args.kwargs["path"] == "/api/val"
            assert len(mock.call_args.kwargs["detail"]) == 1
            assert mock.call_args.kwargs["detail"][0]["loc"] == ["query", "foo"]


def test_middleware_request_logging(app: Starlette, mocker: MockerFixture):
    from starlette.testclient import TestClient

    from apitally.client.request_logging import BODY_TOO_LARGE

    mock = mocker.patch("apitally.client.request_logging.RequestLogger.log_request")

    with TestClient(app) as client:
        response = client.get("/api/foo/123?foo=bar", headers={"Test-Header": "test"})
        assert response.status_code == 200
        mock.assert_called_once()
        assert mock.call_args is not None
        assert mock.call_args.kwargs["request"]["method"] == "GET"
        assert mock.call_args.kwargs["request"]["path"] == "/api/foo/{bar}"
        assert mock.call_args.kwargs["request"]["url"] == "http://testserver/api/foo/123?foo=bar"
        assert ("test-header", "test") in mock.call_args.kwargs["request"]["headers"]
        assert mock.call_args.kwargs["request"]["consumer"] == "test"
        assert mock.call_args.kwargs["response"]["status_code"] == 200
        assert mock.call_args.kwargs["response"]["response_time"] > 0
        assert ("content-type", "text/plain; charset=utf-8") in mock.call_args.kwargs["response"]["headers"]
        assert mock.call_args.kwargs["response"]["size"] > 0
        assert mock.call_args.kwargs["response"]["body"] == b"foo: 123"

        response = client.post("/api/bar", content=b"foo")
        assert response.status_code == 200
        assert mock.call_count == 2
        assert mock.call_args is not None
        assert mock.call_args.kwargs["request"]["method"] == "POST"
        assert mock.call_args.kwargs["request"]["path"] == "/api/bar"
        assert mock.call_args.kwargs["request"]["url"] == "http://testserver/api/bar"
        assert mock.call_args.kwargs["request"]["body"] == b"foo"
        assert mock.call_args.kwargs["response"]["body"] == b"bar: foo"

        mocker.patch("apitally.starlette.MAX_BODY_SIZE", 2)
        response = client.post("/api/bar", content=b"foo")
        assert response.status_code == 200
        assert mock.call_count == 3
        assert mock.call_args is not None
        assert mock.call_args.kwargs["request"]["body"] == BODY_TOO_LARGE
        assert mock.call_args.kwargs["response"]["body"] == BODY_TOO_LARGE


def test_get_startup_data(app: Starlette, mocker: MockerFixture):
    from apitally.starlette import _get_startup_data

    mocker.patch("apitally.starlette.ApitallyClient")
    if app.middleware_stack is None:
        app.middleware_stack = app.build_middleware_stack()

    data = _get_startup_data(app=app.middleware_stack, app_version="1.2.3", openapi_url=None)
    assert len(data["paths"]) == 7
    assert {"method": "get", "path": "/api/foo"} in data["paths"]
    assert {"method": "post", "path": "/test/task"} in data["paths"]
    assert {"method": "get", "path": "/stream"} in data["paths"]
    assert data["versions"]["starlette"]
    assert data["versions"]["app"] == "1.2.3"
    assert data["client"] == "python:starlette"



from __future__ import annotations

from importlib.util import find_spec
from typing import Optional

import pytest
from pytest_mock import MockerFixture

from .constants import CLIENT_ID, ENV


if find_spec("blacksheep") is None:
    pytest.skip("blacksheep is not available", allow_module_level=True)
else:
    # Need to import these at package level to avoid NameError in BlackSheep
    from blacksheep import Application, Request, Response


@pytest.fixture(scope="module")
async def app(module_mocker: MockerFixture) -> Application:
    module_mocker.patch("apitally.client.client_asyncio.ApitallyClient._instance", None)
    module_mocker.patch("apitally.client.client_asyncio.ApitallyClient.start_sync_loop")
    module_mocker.patch("apitally.client.client_asyncio.ApitallyClient.set_startup_data")
    app = get_app()
    await app.start()
    return app


def get_app() -> Application:
    from blacksheep import StreamedContent, get, post, text

    from apitally.blacksheep import ApitallyConsumer, RequestLoggingConfig, use_apitally

    def identify_consumer(request: Request) -> Optional[ApitallyConsumer]:
        return ApitallyConsumer("test", name="Test")

    app = Application(show_error_details=True)

    use_apitally(
        app,
        client_id=CLIENT_ID,
        env=ENV,
        request_logging_config=RequestLoggingConfig(
            enabled=True,
            log_request_body=True,
            log_response_body=True,
        ),
        app_version="1.2.3",
        identify_consumer_callback=identify_consumer,
    )

    @get("/api/foo")
    def foo() -> Response:
        return text("foo")

    @get("/api/foo/{bar}")
    def foo_bar(bar: str) -> Response:
        return text(f"foo: {bar}")

    @post("/api/bar")
    async def bar(request: Request) -> Response:
        body = await request.text()
        return text("bar: " + body)

    @post("/api/baz")
    def baz():
        raise ValueError("baz")

    @get("/api/stream")
    async def stream() -> Response:
        async def stream_response():
            yield b"foo"
            yield b"bar"

        return Response(200, content=StreamedContent(b"text/plain", stream_response))

    return app


async def test_middleware_requests_ok(app: Application, mocker: MockerFixture):
    from blacksheep.testing import TestClient

    mock = mocker.patch("apitally.client.requests.RequestCounter.add_request")
    client = TestClient(app)

    response = await client.get("/api/foo")
    assert response.status == 200
    mock.assert_called_once()
    assert mock.call_args is not None
    assert mock.call_args.kwargs["consumer"] == "test"
    assert mock.call_args.kwargs["method"] == "GET"
    assert mock.call_args.kwargs["path"] == "/api/foo"
    assert mock.call_args.kwargs["status_code"] == 200
    assert mock.call_args.kwargs["response_time"] > 0

    response = await client.get("/api/foo/123")
    assert response.status == 200
    assert mock.call_count == 2
    assert mock.call_args is not None
    assert mock.call_args.kwargs["path"] == "/api/foo/{bar}"

    response = await client.post("/api/bar")
    assert response.status == 200
    assert mock.call_count == 3
    assert mock.call_args is not None
    assert mock.call_args.kwargs["method"] == "POST"

    response = await client.get("/api/stream")
    assert response.status == 200
    assert mock.call_count == 4
    assert mock.call_args is not None
    assert mock.call_args.kwargs["response_size"] == 6


async def test_middleware_requests_error(app: Application, mocker: MockerFixture):
    from blacksheep.testing import TestClient

    mock1 = mocker.patch("apitally.client.requests.RequestCounter.add_request")
    mock2 = mocker.patch("apitally.client.server_errors.ServerErrorCounter.add_server_error")
    client = TestClient(app)

    response = await client.post("/api/baz")
    assert response.status == 500
    mock1.assert_called_once()
    assert mock1.call_args is not None
    assert mock1.call_args.kwargs["method"] == "POST"
    assert mock1.call_args.kwargs["path"] == "/api/baz"
    assert mock1.call_args.kwargs["status_code"] == 500
    assert mock1.call_args.kwargs["response_time"] > 0

    mock2.assert_called_once()
    assert mock2.call_args is not None
    exception = mock2.call_args.kwargs["exception"]
    assert isinstance(exception, ValueError)


async def test_middleware_requests_unhandled(app: Application, mocker: MockerFixture):
    from blacksheep.testing import TestClient

    mock = mocker.patch("apitally.client.requests.RequestCounter.add_request")
    client = TestClient(app)

    response = await client.post("/xxx")
    assert response.status == 404
    mock.assert_not_called()


async def test_middleware_request_logging(app: Application, mocker: MockerFixture):
    from blacksheep.testing import TestClient, TextContent

    from apitally.client.request_logging import BODY_TOO_LARGE

    mock = mocker.patch("apitally.client.request_logging.RequestLogger.log_request")
    client = TestClient(app)

    response = await client.get("/api/foo/123", query="foo=bar", headers={"test-header": "test"})
    assert response.status == 200
    assert await response.text() == "foo: 123"
    mock.assert_called_once()
    assert mock.call_args is not None
    assert mock.call_args.kwargs["request"]["method"] == "GET"
    assert mock.call_args.kwargs["request"]["path"] == "/api/foo/{bar}"
    assert mock.call_args.kwargs["request"]["url"] == "http://127.0.0.1:8000/api/foo/123?foo=bar"
    assert ("test-header", "test") in mock.call_args.kwargs["request"]["headers"]
    assert mock.call_args.kwargs["request"]["consumer"] == "test"
    assert mock.call_args.kwargs["response"]["status_code"] == 200
    assert mock.call_args.kwargs["response"]["response_time"] > 0
    assert mock.call_args.kwargs["response"]["size"] > 0
    assert mock.call_args.kwargs["response"]["body"] == b"foo: 123"

    response = await client.post("/api/bar", content=TextContent("foo"))
    assert response.status == 200
    assert mock.call_count == 2
    assert mock.call_args is not None
    assert mock.call_args.kwargs["request"]["method"] == "POST"
    assert mock.call_args.kwargs["request"]["path"] == "/api/bar"
    assert mock.call_args.kwargs["request"]["url"] == "http://127.0.0.1:8000/api/bar"
    assert mock.call_args.kwargs["request"]["body"] == b"foo"
    assert mock.call_args.kwargs["response"]["body"] == b"bar: foo"

    mocker.patch("apitally.blacksheep.MAX_BODY_SIZE", 2)
    response = await client.post("/api/bar", content=TextContent("foo"), headers={"Content-Length": "3"})
    assert response.status == 200
    assert mock.call_count == 3
    assert mock.call_args is not None
    assert mock.call_args.kwargs["request"]["body"] == BODY_TOO_LARGE


def test_get_startup_data(app: Application, mocker: MockerFixture):
    from apitally.blacksheep import _get_startup_data

    mocker.patch("apitally.blacksheep.ApitallyClient")

    data = _get_startup_data(app, app_version="1.2.3")
    assert len(data["paths"]) == 5
    assert {"method": "GET", "path": "/api/foo"} in data["paths"]
    assert {"method": "GET", "path": "/api/foo/{bar}"} in data["paths"]
    assert {"method": "POST", "path": "/api/bar"} in data["paths"]
    assert {"method": "POST", "path": "/api/baz"} in data["paths"]
    assert {"method": "GET", "path": "/api/stream"} in data["paths"]
    assert data["versions"]["blacksheep"]
    assert data["versions"]["app"] == "1.2.3"
    assert data["client"] == "python:blacksheep"



from __future__ import annotations

import logging
import random
import time
from contextlib import suppress
from functools import partial
from io import BufferedReader
from queue import Queue
from threading import Event, Thread
from typing import Any, Callable, Dict, Optional
from uuid import UUID

import backoff
import requests

from apitally.client.client_base import MAX_QUEUE_TIME, REQUEST_TIMEOUT, ApitallyClientBase
from apitally.client.logging import get_logger
from apitally.client.request_logging import RequestLoggingConfig


logger = get_logger(__name__)
retry = partial(
    backoff.on_exception,
    backoff.expo,
    requests.RequestException,
    max_tries=3,
    logger=logger,
    giveup_log_level=logging.WARNING,
)


# Function to register an on-exit callback for both Python and IPython runtimes
try:

    def register_exit(func: Callable[..., Any], *args, **kwargs) -> Callable[..., Any]:  # pragma: no cover
        def callback():
            func()
            ipython.events.unregister("post_execute", callback)

        ipython.events.register("post_execute", callback)
        return func

    ipython = get_ipython()  # type: ignore
except NameError:
    from atexit import register as register_exit  # type: ignore[assignment]


class ApitallyClient(ApitallyClientBase):
    def __init__(
        self,
        client_id: str,
        env: str,
        request_logging_config: Optional[RequestLoggingConfig] = None,
        proxy: Optional[str] = None,
    ) -> None:
        super().__init__(client_id=client_id, env=env, request_logging_config=request_logging_config)
        self.proxies = {"https": proxy} if proxy else None
        self._thread: Optional[Thread] = None
        self._stop_sync_loop = Event()
        self._sync_data_queue: Queue[Dict[str, Any]] = Queue()

    def start_sync_loop(self) -> None:
        self._stop_sync_loop.clear()
        if self._thread is None or not self._thread.is_alive():
            self._thread = Thread(target=self._run_sync_loop, daemon=True)
            self._thread.start()
            register_exit(self.stop_sync_loop)

    def _run_sync_loop(self) -> None:
        try:
            last_sync_time = 0.0
            while not self._stop_sync_loop.is_set():
                try:
                    self.request_logger.write_to_file()
                except Exception:  # pragma: no cover
                    logger.exception("An error occurred while writing request logs")

                now = time.time()
                if (now - last_sync_time) >= self.sync_interval:
                    try:
                        with requests.Session() as session:
                            if not self._startup_data_sent:
                                self.send_startup_data(session)
                            self.send_sync_data(session)
                            self.send_log_data(session)
                        last_sync_time = now
                    except Exception:  # pragma: no cover
                        logger.exception("An error occurred during sync with Apitally hub")

                self.request_logger.maintain()
                time.sleep(1)
        finally:
            # Send any remaining data before exiting
            with requests.Session() as session:
                self.send_sync_data(session)
                self.send_log_data(session)

    def stop_sync_loop(self) -> None:
        self._stop_sync_loop.set()
        if self._thread is not None:
            self._thread.join()
            self._thread = None

    def set_startup_data(self, data: Dict[str, Any]) -> None:
        self._startup_data_sent = False
        self._startup_data = self.add_uuids_to_data(data)

    def send_startup_data(self, session: requests.Session) -> None:
        if self._startup_data is not None:
            self._send_startup_data(session, self._startup_data)

    def send_sync_data(self, session: requests.Session) -> None:
        data = self.get_sync_data()
        self._sync_data_queue.put_nowait(data)

        i = 0
        while not self._sync_data_queue.empty():
            data = self._sync_data_queue.get_nowait()
            try:
                if time.time() - data["timestamp"] <= MAX_QUEUE_TIME:
                    if i > 0:
                        time.sleep(random.uniform(0.1, 0.5))
                    self._send_sync_data(session, data)
                    i += 1
            except requests.RequestException:
                self._sync_data_queue.put_nowait(data)
                break
            finally:
                self._sync_data_queue.task_done()

    def send_log_data(self, session: requests.Session) -> None:
        self.request_logger.rotate_file()
        i = 0
        while log_file := self.request_logger.get_file():
            if i > 0:
                time.sleep(random.uniform(0.1, 0.3))
            try:
                with log_file.open_compressed() as fp:
                    self._send_log_data(session, log_file.uuid, fp)
                log_file.delete()
            except requests.RequestException:
                self.request_logger.retry_file_later(log_file)
                break
            i += 1
            if i >= 10:
                break

    @retry(raise_on_giveup=False)
    def _send_startup_data(self, session: requests.Session, data: Dict[str, Any]) -> None:
        logger.debug("Sending startup data to Apitally hub")
        response = session.post(
            url=f"{self.hub_url}/startup",
            json=data,
            timeout=REQUEST_TIMEOUT,
            proxies=self.proxies,
        )
        self._handle_hub_response(response)
        self._startup_data_sent = True
        self._startup_data = None

    @retry()
    def _send_sync_data(self, session: requests.Session, data: Dict[str, Any]) -> None:
        logger.debug("Synchronizing data with Apitally hub")
        response = session.post(
            url=f"{self.hub_url}/sync",
            json=data,
            timeout=REQUEST_TIMEOUT,
            proxies=self.proxies,
        )
        self._handle_hub_response(response)

    def _send_log_data(self, session: requests.Session, uuid: UUID, fp: BufferedReader) -> None:
        logger.debug("Streaming request log data to Apitally hub")
        response = session.post(
            url=f"{self.hub_url}/log?uuid={uuid}",
            data=fp,
            timeout=REQUEST_TIMEOUT,
            proxies=self.proxies,
        )
        if response.status_code == 402 and "Retry-After" in response.headers:
            with suppress(ValueError):
                retry_after = int(response.headers["Retry-After"])
                self.request_logger.suspend_until = time.time() + retry_after
                self.request_logger.clear()
                return
        self._handle_hub_response(response)

    def _handle_hub_response(self, response: requests.Response) -> None:
        if response.status_code == 404:
            self.enabled = False
            self.stop_sync_loop()
            logger.error("Invalid Apitally client ID: %s", self.client_id)
        elif response.status_code == 422:
            logger.error("Received validation error from Apitally hub: %s", response.json())
        else:
            response.raise_for_status()
